import json
import requests

class TFEnterpriseVariableAPI():
    """ Terraform Enterprise Variable API wrapper """
    def __init__(self, token, org, workspace, url):
        self.url                      = url
        self.headers                  = {}
        self.headers["Authorization"] = "Bearer %s" % (token)
        self.headers["Content-Type"]  = "application/vnd.api+json"
        self.workspace_id             = self.get_workspace_id(org, workspace)

    def get_workspace_id(self, org, workspace):
        # NOTE: consider caching the results from this to minimize future requests
        workspaces = requests.get("%s/api/v2/organizations/%s/workspaces" % (self.url, org), headers=self.headers).json()
        for ws in workspaces['data']:
            if workspace == ws['attributes']['name']:
                return ws['id']
        return False

    def get_variables(self):
        """ returns a list of variables in the defined workspace. """
        return requests.get("%s/api/v2/workspaces/%s/vars" % (self.url, self.workspace_id), headers=self.headers).json()['data']

    # TODO: consider implementing
    # def get_variable(self):
    #     pass

    def create_variable(self, var):
        payload = {}
        payload['data'] = {}
        payload['data']['type'] = "vars"
        payload['data']['attributes'] = var
        r = requests.post("%s/api/v2/workspaces/%s/vars" % (self.url, self.workspace_id), headers=self.headers, json=payload)

    def update_variable(self, id, var):
        payload = {}
        payload['data'] = {}
        payload['data']['type'] = "vars"
        payload['data']['attributes'] = var
        r = requests.patch("%s/api/v2/workspaces/%s/vars/%s" % (self.url, self.workspace_id, id), headers=self.headers, json=payload)

    def delete_variable(self, id):
        r = requests.delete("%s/api/v2/workspaces/%s/vars/%s" % (self.url, self.workspace_id, id), headers=self.headers)


import os
import yaml
from typing import Any, IO
from cerberus import Validator

class VarException(Exception):
    pass

class Loader(yaml.SafeLoader):
    """Customized YAML Loader"""

    def __init__(self, stream: IO) -> None:
        """Initialise Custom YAML Loader."""

        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.path.curdir

        super().__init__(stream)

class VarFileManager:
    """ variable manager class """
    def __init__(self, f):
        self._set_vfile_schema()
        yaml.add_constructor('!env', self._env, Loader)
        with open(f) as vfile:
            self.vars = yaml.load(vfile, Loader=Loader)
        self._validate_vars()
    
    def _env(self, loader: Loader, node: yaml.Node) -> Any:
        """
        _env is a custom method passed to the yaml Loader to provide
        the ability to load yaml values from environment variables.
        """
        if node.value in os.environ:
            return os.environ[node.value]
        raise VarException("Undefined environment variable '%s' referenced in provided var file." % (node.value))
    
    def _validate_vars(self):
        """
        _validate_vars will validate the vars set by the provided var file according to the defined var file schema.
        """
        self.validator = Validator(self.varfile_schema)
        if not self.validator.validate(self.vars):
            error_msg = "The provided var file has the following errors: \n"
            # print(self.validator.errors)
            for k in self.validator.errors:
                if k == 'variables':
                    # error_msg += "\n%s: %s\n" % (k, self.validator.errors[k][0])
                    error_msg += "\n%s:\n" % (k)
                    for msg in self.validator.errors[k]:
                        for idx in msg:
                            for err in msg[idx]:
                                for k in err:
                                    error_msg += "%s) %s: %s\n" % (idx + 1, k, err[k][0])
                else:
                    error_msg += "\n%s: %s\n" % (k, self.validator.errors[k][0])
            
            raise VarException(error_msg)

    def _set_vfile_schema(self):
        """ 
        _set_vfile_schema is an internal private method that simply defines the validation schema of a var file 
        """
        self.varfile_schema = {
            'url': {
                'type': 'string',
                'required': False
            },
            'envtoken': {
                'type': 'string',
                'required': True
            },
            'organization': {
                'type': 'string',
                'required': True
            },
            'workspace': {
                'type': 'string',
                'required': True
            },
            'variables': {
                'type': 'list',
                'required': True,
                'nullable': True,
                'schema': {
                    'type': 'dict',
                    'schema': {
                        'key': {
                            'type': 'string',
                            'required': True
                        },
                        'value': {
                            'type': 'string',
                            'required': True
                        },
                        'category': {
                            'type': 'string',
                            'allowed': ['env', 'terraform'],
                            'default': 'terraform'
                        },
                        'description': {
                            'type': 'string'
                        },
                        'hcl': {
                            'type': 'boolean',
                            'default': False
                        },
                        'sensitive': {
                            'type': 'boolean',
                            'default': False
                        }
                    }

                },
            }
        }


import os
import sys
import click
import json
from os import path
from .utils.varfile import VarFileManager, VarException
from .utils.api import TFEnterpriseVariableAPI
from tabulate import tabulate

@click.group()
def cli():
  pass

@cli.command()
@click.argument('org')
@click.argument('workspace')
@click.argument('envtoken')
@click.option('--url', 'url')
@click.option('--alias', 'alias')
def init(org, workspace, envtoken, url, alias):
  """ 
  initialize cli to work with a specific workspace\n
  ORG:        organization name\n
  ENVTOKEN:   environment variable to fetch api token from\n
  WORKSPACE:  name of the workspace to select for the current session\n
  URL:        option to provide self hosted Terraform Cloud url\n
  ALIAS:      option to alias the workspace name for CLI use.

  """
  if not envtoken in os.environ:
    print("VARIABLE ERROR: '%s' is not an exported environment variable" % (envtoken))
    sys.exit(1)
  
  session = {
    'org':       org,
    'url':       url or "https://app.terraform.io",
    'token':     os.environ[envtoken],
    'workspace': workspace
  }
  
  # write the session to the /tmp folder for the given workspace
  # TODO: add support for windows environments?
  session_path = "%s/%s.json" % ('/tmp', workspace)
  if alias:
    session_path = "%s/%s.json" % ('/tmp', alias)
  with open(session_path, 'w+') as session_file:
    json.dump(session, session_file)
  
  # verify file exist, and return a message indicating status of initialization
  if path.exists(session_path):
    print("Initalized!")
    sys.exit(0)
  else:
    print("Session initialization failed.")
    sys.exit(0)

# @cli.command()
# def add():
#   """ add a variable to the workspace """
#   pass

# @cli.command()
# def rem():
#   """ remove a variable from the workspace """
#   pass

@cli.command()
@click.argument('WORKSPACE')
def show(workspace):
  """ show variables in the workspace """
  session_path = "%s/%s.json" % ('/tmp', workspace)
  if not path.isfile(session_path):
    print("INIT ERROR: The %s workspace has not been initialized." % (workspace))
    sys.exit(1)
  
  # load the cached session
  with open(session_path, 'r') as session_file:
    session = json.load(session_file)

  # get the variables by making an api call
  api = TFEnterpriseVariableAPI(**session)
  
  variables = api.get_variables()
  headers = ['Key', 'Value', 'Sensitive', 'Type', 'HCL']
  envdata = []
  tfvdata = []
  for k in variables:
      if k['attributes']['category'] == 'env':
          envdata.append([
              k['attributes']['key'],
              k['attributes']['value'],
              k['attributes']['sensitive'],
              k['attributes']['category'],
              k['attributes']['hcl']
          ])
      if k['attributes']['category'] == 'terraform':
          tfvdata.append([
              k['attributes']['key'],
              k['attributes']['value'],
              k['attributes']['sensitive'],
              k['attributes']['category'],
              k['attributes']['hcl']
          ])

  print("TERRAFORM VARIABLES:\n")
  print(tabulate(tfvdata, headers=headers))
  print("\n")
  print("ENVIRONMENT VARIABLES:\n")
  print(tabulate(envdata, headers=headers))
  sys.exit(0)

@cli.command()
@click.argument('varfile')
def sync(varfile):
  """ sync variables from a var file to the workspace """
  # load varfile
  vfile = VarFileManager(varfile)
  url = "https://app.terraform.io"
  if 'url' in vfile.vars:
    url = vfile.vars['url']

  session = {
    'org':       vfile.vars['organization'],
    'url':       url,
    'token':     os.environ[vfile.vars['envtoken']],
    'workspace': vfile.vars['workspace']
  }

  api = TFEnterpriseVariableAPI(**session)
  variables = api.get_variables()

  envvars = {}
  tfvvars = {}

  if vfile.vars['variables'] is None:
      vfile.vars['variables'] = []

  for k in variables:
      if k['attributes']['category'] == 'env':
          envvars[k['attributes']['key']] = k['id']
      if k['attributes']['category'] == 'terraform':
          tfvvars[k['attributes']['key']] = k['id']
  
  # # payload['data']['type'] = "vars"
  for var in vfile.vars['variables']:
      if 'category' not in var.keys():
          var['category'] = 'terraform'
      if 'description' not in var.keys():
          var['description'] = ''
      if 'hcl' not in var.keys():
          var['hcl'] = False
      if 'sensitive' not in var.keys():
          var['sensitive'] = False

      if var['key'] in envvars.keys() and var['category'] == 'env':
          # we need to update the variable
          print("UPDATING ENV %s ..." % (var['key']))
          api.update_variable(envvars[var['key']], var)
          del envvars[var['key']]
      
      elif var['key'] in tfvvars.keys() and var['category'] == 'terraform':
          # we need to update the variable
          print("UPDATING TF %s ..." % (var['key']))
          api.update_variable(tfvvars[var['key']], var)
          del tfvvars[var['key']]
      
      elif var['key'] not in envvars and var['key'] not in tfvvars:
          # we need to create the variable
          print("CREATING %s ..." % var['key'])
          api.create_variable(var)
      else:
          print("shouldn't get here..")
          pass
  
  for var in envvars:
      print("REMOVING %s ..." % (var))
      api.delete_variable(envvars[var])
  
  for var in tfvvars:
      print("REMOVING %s ..." % (var))
      api.delete_variable(tfvvars[var])

def main():
    """Terraform Enterprise Variable Management CLI"""
    try:
      cli()
    except VarException as varerr:
      print("VARIABLE ERROR: %s" % (str(varerr)))
      sys.exit(1)
    except Exception as err:
      print("UNKNOWN ERROR: %s" % (str(err)))
      sys.exit(1)


"""
Setup script for tfvarman
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.
version = "0.0.0-4" #NOTE: please blame pypy for the weird version numbers...

setup(
    name='tfvarman',
    version=version,
    description="variable management cli tool for terraform cloud",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/dsikes/tfvarman',
    author='Dan Sikes',
    author_email='dansikes7@gmail.com',
    keywords='terraform, automation, tfvars, terraform variables',
    packages=[
        'tfvarman',
        'tfvarman.utils'
    ],
    install_requires=[
        'click',
        'cerberus',
        'tabulate',
        'pyaml',
        'requests'
    ],
    entry_points = {
        'console_scripts': ['tfvarman=tfvarman.cli:main'],
    },
    
    project_urls={
        'Source': 'https://github.com/dsikes/tfvarman',
    },
)